@@ -1,3 +1,4 @@
+from turtle import color
 import numpy as np
 import pandas as pd
 from scipy import integrate, signal, stats
@@ -26,28 +27,48 @@ def clean_df(df, mask=None):
 
     return df_cleaned
 
-class WhiteNoise:
-    ''' A PSM that adds white noise to the input climate signal.
+class TempPlusNoise:
+    ''' A PSM that adds noise to the input climate signal.
+
+    Args:
+        pobj (cfr.proxy.ProxyRecord): the proxy record object
+        climate_required (cfr.climate.ClimateField): the required climate field object for running this PSM
     '''
-    def __init__(self, pobj=None, climate_required=[]):
+    def __init__(self, pobj=None, climate_required=['tas']):
         self.pobj = pobj
         self.climate_required = climate_required
 
-    def calibrate(self, SNR=10, seasonality=list(range(13))):
-        sigma = np.std(self.pobj.value) / SNR
-        self.wn = np.random.normal(0, sigma, np.size(self.pobj.value))
+    def calibrate(self, SNR=10, seasonality=list(range(1, 13)), vn='model.tas',
+                  seed=None, noise='white', colored_noise_kws=None):
+        ''' Calibrate the PSM.'''
+        sigma = np.nanstd(self.pobj.clim[vn].da.values) / SNR
+        if noise == 'white':
+            rng = np.random.default_rng(seed)
+            self.noise = rng.normal(0, sigma, np.size(self.pobj.clim[vn].da.values))
+        elif noise == 'colored':
+            colored_noise_kws = {} if colored_noise_kws is None else colored_noise_kws
+            _colored_noise_kws = {'seed': seed, 'alpha': 1, 't': self.pobj.clim[vn].time}
+            _colored_noise_kws.update(colored_noise_kws)
+            self.noise = utils.colored_noise(**_colored_noise_kws)
+
         calib_details = {
-            'PSMmse': np.mean(self.wn**2),
+            'PSMmse': np.mean(self.noise**2),
             'SNR': SNR,
             'seasonality': seasonality,
         }
         self.calib_details = calib_details
 
-    def forward(self):
+    def forward(self, vn='model.tas', no_noise=False):
+        ''' Forward the PSM.'''
+        if no_noise:
+            value = self.pobj.clim[vn].da.values 
+        else:
+            value=self.pobj.clim[vn].da.values + self.noise
+
         pp = ProxyRecord(
-            pid=f'pseudo_{self.pobj.pid}',
-            time=self.pobj.time,
-            value=self.pobj.value + self.wn,
+            pid=self.pobj.pid,
+            time=self.pobj.clim[vn].time,
+            value=value,
             lat=self.pobj.lat,
             lon=self.pobj.lon,
             ptype=self.pobj.ptype,
@@ -61,12 +82,18 @@ class WhiteNoise:
         return pp
 
 class Linear:
-    def __init__(self, pobj=None, climate_required = ['tas']):
+    ''' A PSM that is based on univariate linear regression.
+
+    Args:
+        pobj (cfr.proxy.ProxyRecord): the proxy record object
+        climate_required (cfr.climate.ClimateField): the required climate field object for running this PSM
+    '''
+    def __init__(self, pobj=None, climate_required=['tas']):
         self.pobj = pobj
         self.climate_required = climate_required
 
     def calibrate(self, calib_period=None, nobs_lb=25, metric='fitR2adj',
-        season_list=[list(range(1, 13))], annualize_exog=True, exog_name='obs_tas', **fit_args):
+        season_list=[list(range(1, 13))], annualize_exog=True, exog_name='obs.tas', **fit_args):
         exog = self.pobj.clim[exog_name]
 
         if type(season_list[0]) is not list:
@@ -135,7 +162,7 @@ class Linear:
             self.calib_details = None
             self.model = None
 
-    def forward(self, exog_name='model_tas'):
+    def forward(self, exog_name='model.tas'):
         sn = self.calib_details['seasonality']
         exog = self.pobj.clim[exog_name]
         if sn != 'monthly':
@@ -146,7 +173,7 @@ class Linear:
         }
 
         pp = ProxyRecord(
-            pid=f'pseudo_{self.pobj.pid}',
+            pid=self.pobj.pid,
             time=exog.time,
             value=np.array(self.model.predict(exog=exog_dict).values),
             lat=self.pobj.lat,
@@ -163,13 +190,19 @@ class Linear:
 
 
 class Bilinear:
+    ''' A PSM that is based on bivariate linear regression.
+
+    Args:
+        pobj (cfr.proxy.ProxyRecord): the proxy record object
+        climate_required (cfr.climate.ClimateField): the required climate field object for running this PSM
+    '''
     def __init__(self, pobj=None, climate_required = ['tas', 'pr']):
         self.pobj = pobj
         self.climate_required = climate_required
 
     def calibrate(self, calib_period=None, nobs_lb=25, metric='fitR2adj',
         season_list1=[list(range(1, 13))], season_list2=[list(range(1, 13))],
-        annualize_exog=True, exog1_name='obs_tas', exog2_name='obs_pr', **fit_args):
+        annualize_exog=True, exog1_name='obs.tas', exog2_name='obs.pr', **fit_args):
         exog1 = self.pobj.clim[exog1_name]
         exog2 = self.pobj.clim[exog2_name]
 
@@ -246,7 +279,7 @@ class Bilinear:
             self.calib_details = None
             self.model = None
 
-    def forward(self, exog1_name='model_tas', exog2_name='model_pr'):
+    def forward(self, exog1_name='model.tas', exog2_name='model.pr'):
         exog1 = self.pobj.clim[exog1_name]
         exog2 = self.pobj.clim[exog2_name]
         sn1, sn2 = self.calib_details['seasonality']
@@ -263,7 +296,7 @@ class Bilinear:
         }
 
         pp = ProxyRecord(
-            pid=f'pseudo_{self.pobj.pid}',
+            pid=self.pobj.pid,
             time=exog1.time,
             value=np.array(self.model.predict(exog=exog_dict).values),
             lat=self.pobj.lat,
@@ -280,7 +313,13 @@ class Bilinear:
 
 
 class Ice_d18O():
-    def __init__(self, pobj=None, tas_name='model_tas', pr_name='model_pr', psl_name='model_psl', d18O_name='model_d18O',
+    ''' The ice core d18O model adopted from PRYSM (https://github.com/sylvia-dee/PRYSM)
+
+    Args:
+        pobj (cfr.proxy.ProxyRecord): the proxy record object
+        climate_required (cfr.climate.ClimateField): the required climate field object for running this PSM
+    '''
+    def __init__(self, pobj=None, tas_name='model.tas', pr_name='model.pr', psl_name='model.psl', d18O_name='model.d18O',
                  climate_required=['tas', 'pr', 'psl', 'd18O']):
         self.pobj = pobj
         self.tas_name = tas_name
@@ -635,7 +674,7 @@ class Ice_d18O():
         ice_diffused = ice_archive(d18O_ice, pr_ann.da.values, tas_ann.da.values, psl_ann.da.values, nproc=nproc)
 
         pp = ProxyRecord(
-            pid=f'pseudo_{self.pobj.pid}',
+            pid=self.pobj.pid,
             time=tas_ann.time,
             value=ice_diffused[::-1],
             lat=self.pobj.lat,
@@ -650,11 +689,11 @@ class Ice_d18O():
         return pp
 
 class Lake_VarveThickness():
-    ''' The varve thickness model
+    ''' The varve thickness model.
 
     It takes summer temperature as input (JJA for NH and DJF for SH).
     '''
-    def __init__(self, pobj=None, model_tas_name='model_tas', H=None, shape=None, mean=None, SNR=None, seed=None,
+    def __init__(self, pobj=None, model_tas_name='model.tas', H=None, shape=None, mean=None, SNR=None, seed=None,
                  climate_required=['tas']):
         self.pobj = pobj
         self.H = H
@@ -757,7 +796,7 @@ class Lake_VarveThickness():
         varve_res = simpleVarveModel(tas_ann.da.values, **kwargs)
 
         pp = ProxyRecord(
-            pid=f'pseudo_{self.pobj.pid}',
+            pid=self.pobj.pid,
             time=tas_ann.time,
             value=np.array(varve_res['varves'])[0],
             lat=self.pobj.lat,
@@ -774,7 +813,7 @@ class Lake_VarveThickness():
 class Coral_SrCa:
     ''' The coral Sr/Ca model
     '''
-    def __init__(self, pobj=None, model_tos_name='model_tos', b=10.553, a=None, seed=None, climate_required='tos'):
+    def __init__(self, pobj=None, model_tos_name='model.tos', b=10.553, a=None, seed=None, climate_required='tos'):
         self.pobj = pobj
         self.model_tos_name = model_tos_name
         self.b = b
@@ -796,7 +835,7 @@ class Coral_SrCa:
         SrCa = self.a*self.pobj.clim[self.model_tos_name].da.values + self.b
 
         pp = ProxyRecord(
-            pid=f'pseudo_{self.pobj.pid}',
+            pid=self.pobj.pid,
             time=self.pobj.clim[self.model_tos_name].time,
             value=SrCa,
             lat=self.pobj.lat,
@@ -817,7 +856,7 @@ class Coral_d18O:
        model of coral \u03b418O, Geophys.Res.Lett., 38, L14706, doi:10.1029/2011GL048224.>
        Returns a numpy array that is the same size and shape as the input vectors for SST, SSS.
     '''
-    def __init__(self, pobj=None, model_tos_name='model_tos', model_d18Osw_name='model_d18Osw',
+    def __init__(self, pobj=None, model_tos_name='model.tos', model_d18Osw_name='model.d18Osw',
         climate_required=['tos', 'd18Osw'], species='default',
         b1=0.3007062, b2=0.2619054, b3=0.436509, b4=0.1552032, b5=0.15):
         self.pobj = pobj
@@ -955,7 +994,7 @@ class Coral_d18O:
         ####################
         # run the model
 
-        d18O = pseudocoral(
+        value = pseudocoral(
             self.pobj.clim[self.model_tos_name].da.values,
             d18O=self.pobj.clim[self.model_d18Osw_name].da.values,
             lat=self.pobj.lat,
@@ -969,9 +1008,9 @@ class Coral_d18O:
         )
 
         pp = ProxyRecord(
-            pid=f'pseudo_{self.pobj.pid}',
+            pid=self.pobj.pid,
             time=self.pobj.clim[self.model_tos_name].time,
-            value=d18O,
+            value=value,
             lat=self.pobj.lat,
             lon=self.pobj.lon,
             ptype=self.pobj.ptype,
@@ -988,8 +1027,8 @@ class VSLite:
     ''' The VS-Lite tree-ring width model that takes monthly tas, pr as input.
     '''
     def __init__(self, pobj=None,
-            obs_tas_name='obs_tas', obs_pr_name='obs_pr',
-            model_tas_name='model_tas', model_pr_name='model_pr',
+            obs_tas_name='obs.tas', obs_pr_name='obs.pr',
+            model_tas_name='model.tas', model_pr_name='model.pr',
             climate_required=['tas', 'pr'],
         ):
         self.pobj = pobj
@@ -1075,7 +1114,7 @@ class VSLite:
         )
 
         pp = ProxyRecord(
-            pid=f'pseudo_{self.pobj.pid}',
+            pid=self.pobj.pid,
             time=np.arange(syear, eyear+1),
             value=vsl_res['trw'],
             lat=self.pobj.lat,
